#include <libojmixer/ojmixer.h>

#define CONFIG_MAJOR 0
#define CONFIG_MINOR 1

static gboolean config_loaded = FALSE;

void
oj_config_load(OjMixer *mixer)
{
	char *buffer, *c, *d;
	char *configfile;
	FILE *fp;
	struct stat st;
	OjPreset *preset = NULL;

	configfile = g_build_filename(g_get_home_dir(), ".ojmixerrc", NULL);

	if ((fp = fopen(configfile, "rw")) == NULL)
	{
		config_loaded = TRUE;
		g_free(configfile);
		return;
	}

	g_free(configfile);

	if (fstat(fileno(fp), &st) != 0)
	{
		config_loaded = TRUE;
		fclose(fp);
		return;
	}

	buffer = g_new0(char, st.st_size + 1);

	fread(buffer, 1, st.st_size, fp);

	fclose(fp);

	for (c = buffer; *c != '\0'; c = strchr(c, '\n') + 1)
	{
		if (*c == '#' || *c == '\r' || *c == '\n')
			continue;
		else if (*c == '[')
		{
			d = strchr(c, ']');
			*d = '\0';

			if (preset != NULL)
				oj_mixer_add_preset(mixer, preset);

			preset = oj_preset_new();
			oj_preset_set_name(preset, c + 1);

			c = d + 1;
		}
		else if (*c >= '0' && *c <= '9')
		{
			int device = 0;
			int volume = 0;

			if (preset == NULL)
			{
				fprintf(stderr, _("Junk data found. Aborting.\n"));
				exit(1);
			}

			if (sscanf(c, "%d: %d", &device, &volume) == 2)
				oj_preset_set_volume(preset, device, volume);
		}
		else if (strncmp(c, "rec: ", 5) == 0)
		{
			int device = 0;

			if (preset == NULL)
			{
				fprintf(stderr, _("Junk data found. Aborting.\n"));
				exit(1);
			}

			if (sscanf(c, "rec: %d", &device) == 1)
				oj_preset_set_rec_device(preset, device);
		}
		else if (strncmp(c, "version ", 8) == 0)
		{
			int major, minor;

			if (sscanf(c, "version %d.%d", &major, &minor) == 2)
			{
				if (CONFIG_MAJOR < major ||
					(CONFIG_MAJOR == major && CONFIG_MINOR < minor))
				{
					fprintf(stderr, _("Unsupported config version %d.%d\n"),
							major, minor);
					exit(1);
				}
			}
			else
			{
				fprintf(stderr, _("Junk data found. Aborting.\n"));
				exit(1);
			}
		}
		else if (strncmp(c, "fade_on_switch: ", 16) == 0)
		{
			int value;

			if (sscanf(c, "fade_on_switch: %d", &value) == 1)
				oj_mixer_set_fade_on_switch(mixer, (gboolean)value);
		}
	}

	if (preset != NULL)
		oj_mixer_add_preset(mixer, preset);

	free(buffer);

	config_loaded = TRUE;
}

void
oj_config_save(OjMixer *mixer)
{
	OjPreset *preset;
	GList *l;
	char *configfile;
	FILE *fp;

	g_return_if_fail(mixer != NULL);

	if (!config_loaded)
		return;

	configfile = g_build_filename(g_get_home_dir(), ".ojmixerrc", NULL);

	if ((fp = fopen(configfile, "w")) == NULL)
	{
		g_free(configfile);

		fprintf(stderr, _("Unable to open %s for writing.\n"), configfile);
		return;
	}

	g_free(configfile);

	fprintf(fp, "# Do not edit this file!\n");
	fprintf(fp, "version %d.%d\n\n", CONFIG_MAJOR, CONFIG_MINOR);

	fprintf(fp, "fade_on_switch: %d\n", oj_mixer_get_fade_on_switch(mixer));

	fprintf(fp, "\n\n");

	for (l = mixer->presets; l != NULL; l = l->next)
	{
		int i;

		preset = l->data;

		fprintf(fp, "[%s]\n", oj_preset_get_name(preset));

		fprintf(fp, "rec: %d\n", oj_preset_get_rec_device(preset));

		for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
		{
			if (oj_mixer_is_device_supported(mixer, i))
				fprintf(fp, "%d: %d\n", i, oj_preset_get_volume(preset, i));
		}

		fprintf(fp, "\n");
	}

	fclose(fp);
}
