#include <libojmixer/ojmixer.h>
#include <libojmixer/ojconfig.h>
#include "oj-marshal.h"

#define FADE_SECONDS 2

static void oj_mixer_class_init(GObjectClass *klass, gpointer data);
static void oj_mixer_instance_init(GObject *object, gpointer data);
static void oj_mixer_instance_finalize(GObject *object);

struct OjFadeInfo
{
	OjMixer *mixer;
	OjPreset *preset;

	int devices[SOUND_MIXER_NRDEVICES];

	int end_tick;
	int cur_tick;
};

enum
{
	VOLUME_CHANGED,
	RECSRC_CHANGED,
	PRESET_ADDED,
	PRESET_REMOVED,
	PRESET_SWITCHED,
	LAST_SIGNAL
};

static char *dev_names[SOUND_MIXER_NRDEVICES]  = SOUND_DEVICE_NAMES;
/* static char *dev_labels[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_LABELS; */

static guint signals[LAST_SIGNAL] = { 0 };

static const char *dev_labels[SOUND_MIXER_NRDEVICES] =
{
	_("Volume"),
	_("Bass"),
	_("Treble"),
	_("Synth"),
	_("PCM"),
	_("Speaker"),
	_("Line"),
	_("Mic"),
	_("CD"),
	_("Mix"),
	_("PCM 2"),
	_("Record"),
	_("Input Gain"),
	_("Output Gain"),
	_("Line 1"),
	_("Line 2"),
	_("Line 3"),
	_("Digital1"),
	_("Digital2"),
	_("Digital3"),
	_("Phone In"),
	_("Phone Out"),
	_("Video"),
	_("Radio"),
	_("Monitor")
};

static gboolean
fade_timeout(struct OjFadeInfo *fade)
{
	int i, preset_vol, new_vol;
	double pct1, pct2;

	pct1 = (double)(fade->end_tick - fade->cur_tick) / (double)fade->end_tick;
	pct2 = (double)fade->cur_tick / (double)fade->end_tick;

	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
		if (fade->devices[i] != -1)
		{
			preset_vol = oj_preset_get_volume(fade->preset, i);

			if (preset_vol == fade->devices[i])
				continue;

			new_vol = (fade->devices[i] * pct1) + (preset_vol * pct2);

			oj_mixer_set_volume(fade->mixer, i, new_vol);
		}
	}

	fade->cur_tick++;

	if (fade->cur_tick == fade->end_tick + 1)
	{
		fade->mixer->is_fading = FALSE;

		g_free(fade);

		return FALSE;
	}

	return TRUE;
}

GType
oj_mixer_get_type(void)
{
	static GType oj_mixer_type = 0;

	if (!oj_mixer_type)
	{
		static const GTypeInfo oj_mixer_info =
		{
			sizeof(OjMixerClass),
			NULL,
			NULL,
			(GClassInitFunc)oj_mixer_class_init,
			NULL,
			NULL,
			sizeof(OjMixer),
			0,
			(GInstanceInitFunc)oj_mixer_instance_init
		};

		oj_mixer_type = g_type_register_static(G_TYPE_OBJECT, "OjMixer",
											   &oj_mixer_info, 0);
	}

	return oj_mixer_type;
}

static void
oj_mixer_class_init(GObjectClass *klass, gpointer data)
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS(klass);
	object_class->finalize = oj_mixer_instance_finalize;

	signals[VOLUME_CHANGED] =
		g_signal_new("volume_changed",
					 G_TYPE_FROM_CLASS(object_class),
					 G_SIGNAL_RUN_LAST,
					 G_STRUCT_OFFSET(OjMixerClass, volume_changed),
					 NULL, NULL,
					 oj_marshal_VOID__UINT_UINT,
					 G_TYPE_NONE, 2,
					 G_TYPE_UINT, G_TYPE_UINT);

	signals[RECSRC_CHANGED] =
		g_signal_new("recsrc_changed",
					 G_TYPE_FROM_CLASS(object_class),
					 G_SIGNAL_RUN_LAST,
					 G_STRUCT_OFFSET(OjMixerClass, recsrc_changed),
					 NULL, NULL,
					 g_cclosure_marshal_VOID__UINT,
					 G_TYPE_NONE, 1,
					 G_TYPE_UINT);

	signals[PRESET_ADDED] =
		g_signal_new("preset_added",
					 G_TYPE_FROM_CLASS(object_class),
					 G_SIGNAL_RUN_LAST,
					 G_STRUCT_OFFSET(OjMixerClass, preset_added),
					 NULL, NULL,
					 g_cclosure_marshal_VOID__BOXED,
					 G_TYPE_NONE, 1,
					 TYPE_OJ_PRESET);

	signals[PRESET_REMOVED] =
		g_signal_new("preset_removed",
					 G_TYPE_FROM_CLASS(object_class),
					 G_SIGNAL_RUN_LAST,
					 G_STRUCT_OFFSET(OjMixerClass, preset_removed),
					 NULL, NULL,
					 g_cclosure_marshal_VOID__BOXED,
					 G_TYPE_NONE, 1,
					 TYPE_OJ_PRESET);

	signals[PRESET_SWITCHED] =
		g_signal_new("preset_switched",
					 G_TYPE_FROM_CLASS(object_class),
					 G_SIGNAL_RUN_LAST,
					 G_STRUCT_OFFSET(OjMixerClass, preset_switched),
					 NULL, NULL,
					 g_cclosure_marshal_VOID__BOXED,
					 G_TYPE_NONE, 1,
					 TYPE_OJ_PRESET);
}

static void
oj_mixer_instance_init(GObject *object, gpointer data)
{
}

static void
oj_mixer_instance_finalize(GObject *object)
{
	GObjectClass *parent_class;
	OjMixer *mixer;
	GList *l;

	mixer = OJ_MIXER(object);

	for (l = mixer->presets; l != NULL; l = l->next)
	{
		OjPreset *preset = l->data;

		g_object_unref(preset);
	}

	g_list_free(mixer->presets);

	parent_class = g_type_class_peek_parent(G_OBJECT_GET_CLASS(object));
	parent_class->finalize(object);
}

OjMixer *
oj_mixer_new(void)
{
	return g_object_new(TYPE_OJ_MIXER, NULL);
}

gboolean
oj_mixer_open(OjMixer *mixer, const char *dev)
{
	g_return_val_if_fail(mixer != NULL, FALSE);
	g_return_val_if_fail(dev   != NULL, FALSE);

	if ((mixer->fd = open(dev, O_RDWR)) < 0)
		return FALSE;

	oj_mixer_get_status(mixer);

	if (!mixer->devmask)
		return FALSE;

	return TRUE;
}

int
oj_mixer_get_status(OjMixer *mixer)
{
	g_return_val_if_fail(mixer != NULL, -1);

	if (SOUND_IOCTL(mixer->fd, SOUND_MIXER_READ_DEVMASK, &mixer->devmask) == -1)
		return EREADDEV;

	if (SOUND_IOCTL(mixer->fd, SOUND_MIXER_READ_RECMASK, &mixer->recmask) == -1)
		return EREADRECMASK;

	if (SOUND_IOCTL(mixer->fd, SOUND_MIXER_READ_STEREODEVS,
					&mixer->stereodevs) == -1)
		return EREADSTEREO;

	return 0;
}

int
oj_mixer_get_volume(OjMixer *mixer, int device, int *left,
					int *right, int *volume)
{
	int left_right;
	int l, r;

	g_return_val_if_fail(mixer != NULL, -1);

	if (SOUND_IOCTL(mixer->fd, MIXER_READ(device), &left_right) == -1)
		return EREADMIX;

	l = left_right & 0xFF;
	r = left_right >> 8;

	if (left   != NULL) *left   = l;
	if (right  != NULL) *right  = r;
	if (volume != NULL) *volume = (l + r) / 2;

	return 0;
}

int
oj_mixer_get_rec_source(OjMixer *mixer)
{
	int recsrc, i;

	g_return_val_if_fail(mixer != NULL, -1);

	if (SOUND_IOCTL(mixer->fd, SOUND_MIXER_READ_RECSRC, &recsrc) == -1)
		return -1;

	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
		if ((1 << i) & recsrc)
			return i;
	}

	return -1;
}

int
oj_mixer_get_device_count(OjMixer *mixer)
{
	int i, c = 0;

	g_return_val_if_fail(mixer != NULL, 0);

	for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
	{
		if (oj_mixer_is_device_supported(mixer, i))
			c++;
	}

	return c;
}

gboolean
oj_mixer_is_device_supported(OjMixer *mixer, int device)
{
	g_return_val_if_fail(mixer != NULL, FALSE);

	return (mixer->devmask & (1 << device));
}

gboolean
oj_mixer_can_device_record(OjMixer *mixer, int device)
{
	g_return_val_if_fail(mixer != NULL, FALSE);

	return (mixer->recmask & (1 << device));
}

gboolean
oj_mixer_device_has_stereo(OjMixer *mixer, int device)
{
	g_return_val_if_fail(mixer != NULL, FALSE);

	return (mixer->stereodevs & (1 << device));
}

const char *
oj_mixer_get_device_name(OjMixer *mixer, int device)
{
	g_return_val_if_fail(mixer != NULL, NULL);

	return dev_names[device];
}

const char *
oj_mixer_get_device_label(OjMixer *mixer, int device)
{
	g_return_val_if_fail(mixer != NULL, NULL);

	return dev_labels[device];
}

int
oj_mixer_set_volume(OjMixer *mixer, int device, int volume)
{
	int left_right;
	int left, right;

	g_return_val_if_fail(mixer != NULL, EREADMIX);

	if (oj_mixer_get_volume(mixer, device, &left, &right, NULL) != 0)
		return EREADMIX;

	left  = volume;
	right = volume;

	left  = (left > MAXLEVEL) ? MAXLEVEL : left;
	right = (right > MAXLEVEL) ? MAXLEVEL : right;

	left  = (left < 0) ? 0 : left;
	right = (right < 0) ? 0 : right;

	left_right = (right << 8) | left;

	if (SOUND_IOCTL(mixer->fd, MIXER_WRITE(device), &left_right) == -1)
		return EWRITEMIX;

	g_signal_emit(mixer, signals[VOLUME_CHANGED], 0, device, volume);

	return 0;
}

void
oj_mixer_set_fade_on_switch(OjMixer *mixer, gboolean value)
{
	g_return_if_fail(mixer != NULL);

	mixer->fade_on_switch = value;

	oj_config_save(mixer);
}

gboolean
oj_mixer_get_fade_on_switch(const OjMixer *mixer)
{
	g_return_val_if_fail(mixer != NULL, FALSE);

	return mixer->fade_on_switch;
}

void
oj_mixer_switch_preset(OjMixer *mixer, OjPreset *preset, gboolean fade)
{
	g_return_if_fail(mixer  != NULL);
	g_return_if_fail(preset != NULL);

	int i, volume;

	if (fade)
	{
		struct OjFadeInfo *fadeinfo;

		fadeinfo = g_new0(struct OjFadeInfo, 1);

		fadeinfo->mixer    = mixer;
		fadeinfo->preset   = preset;
		fadeinfo->cur_tick = 0;
		fadeinfo->end_tick = FADE_SECONDS * 10;

		for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
		{
			if (!oj_mixer_is_device_supported(mixer, i))
				fadeinfo->devices[i] = -1;
			else
			{
				oj_mixer_get_volume(mixer, i, NULL, NULL, &volume);

				fadeinfo->devices[i] = volume;
			}
		}

		mixer->is_fading = TRUE;

		g_timeout_add(100, (GSourceFunc)fade_timeout, fadeinfo);
	}
	else
	{
		for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
		{
			if (oj_mixer_is_device_supported(mixer, i))
			{
				int volume;

				volume = oj_preset_get_volume(preset, i);

				oj_mixer_set_volume(mixer, i, volume);
			}
		}	
	}

	g_signal_emit(mixer, signals[PRESET_SWITCHED], 0, preset);
}

int
oj_mixer_set_rec_source(OjMixer *mixer, int recsrc)
{
	g_return_val_if_fail(mixer != NULL, -1);

	if (recsrc == oj_mixer_get_rec_source(mixer))
		return 0;

	recsrc = (1 << recsrc);

	if (SOUND_IOCTL(mixer->fd, SOUND_MIXER_WRITE_RECSRC, &recsrc) == -1)
		return EWRITERECSRC;

	g_signal_emit(mixer, signals[RECSRC_CHANGED], 0, recsrc);

	return 0;
}

void
oj_mixer_add_preset(OjMixer *mixer, OjPreset *preset)
{
	g_return_if_fail(mixer  != NULL);
	g_return_if_fail(preset != NULL);

	mixer->presets = g_list_append(mixer->presets, preset);

	g_signal_emit(mixer, signals[PRESET_ADDED], 0, preset);
}

void
oj_mixer_remove_preset(OjMixer *mixer, OjPreset *preset)
{
	g_return_if_fail(mixer  != NULL);
	g_return_if_fail(preset != NULL);

	mixer->presets = g_list_remove(mixer->presets, preset);

	g_signal_emit(mixer, signals[PRESET_REMOVED], 0, preset);

	g_object_unref(preset);
}

OjPreset *
oj_mixer_find_preset(OjMixer *mixer, const char *name)
{
	GList *l;

	if (mixer == NULL || name == NULL)
		return NULL;

	for (l = mixer->presets; l != NULL; l = l->next)
	{
		OjPreset *preset = l->data;

		if (!strcasecmp(oj_preset_get_name(preset), name))
			return preset;
	}

	return NULL;
}

gboolean
oj_mixer_is_fading(OjMixer *mixer)
{
	if (mixer == NULL)
		return FALSE;

	return mixer->is_fading;
}
