/**
 * @file args.c Parses arguments
 * 
 * $Id: args.c,v 1.1 2003/04/08 20:26:38 chipx86 Exp $
 *
 * @Copyright (C) 2001-2003 The GNUpdate Project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA  02111-1307  USA
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <libojmixer/libojmixer.h>
#include "args.h"
#include "settings.h"

#define INVALID_ARG()  \
	{ \
		fprintf(stderr, _("Invalid argument: %s\n"), argv[i]); \
		exit(1); \
	}

static char
__checkArgTypes(char *opt, int argTypeCount, Argument *argTypes,
				char optIndex, Argument **retArg)
{
	char found = 0;
	char result;
	int i;

	*retArg = NULL;
	
	for (i = 0; i < argTypeCount; i++)
	{
		Argument *arg = &argTypes[i];

		result = 0;

		if (optIndex == 0 && arg->shortOpt != 0)
		{
			result = (*opt == arg->shortOpt ? 1 : 0);
		}
		else if (optIndex == 1 && arg->longOpt != NULL)
		{
			result = (strcmp(opt, arg->longOpt) == 0 ? 1 : 0);
		}
		else if (optIndex == 2 && arg->commandOpt != NULL)
		{
			result = (strcmp(opt, arg->commandOpt) == 0 ? 1 : 0);
		}
		
		if (result == 1)
		{
			if (arg->commandOpt != NULL && getCommand() != CMD_UNSPECIFIED)
			{
				fprintf(stderr,
						_("Only one command may be specified at a time!\n"));
				exit(1);
			}

			found = 1;

			if (arg->type == ARG_FUNCTION && arg->fnc != NULL)
			{
				arg->fnc(arg);
			}
			else if (arg->type == ARG_SETTING)
			{
				if (arg->var == NULL)
				{
					fprintf(stderr,
							(_("arg->var == NULL! Contact the developers.\n")));
					exit(1);
				}
			}
			else if (arg->var != NULL)
			{
				if (arg->type == ARG_INT)
				{
					*((int *)arg->var) = arg->value.integer;
				}
				else if (arg->type == ARG_STRING)
				{
					*((char **)arg->var) = arg->value.string;
				}
				else if (arg->type == ARG_COMMAND)
				{
					*((Command *)arg->var) = arg->value.command;
				}
			}

			*retArg = arg;

			break;
		}
	}

	return found;
}

void
parseArgs(int argc, char **argv, int argTypeCount, Argument *argTypes,
		  int *newArgc, char ***newArgv, char keepUnknownFlags)
{
	int i;
	char found = 0;
	Argument *arg;

	if (argc <= 0 || newArgv == NULL || newArgc == NULL)
	{
		if (newArgc != NULL) *newArgc = 0;
		if (newArgv != NULL) *newArgv = NULL;

		return;
	}

	/*
	 * Allocate the new arrays.
	 *
	 * Allocate enough memory for argc strings. At least one option
	 * is required to be passed on the command line, so this array will
	 * be big enough.
	 */
	*newArgv = (char **)malloc(argc * sizeof(char *));

	*newArgc = 0;

	for (i = 0; i < argc; i++)
	{
		int len = strlen(argv[i]);

		/* Check for '-x' opts */
		if (len >= 2 && argv[i][0] == '-' && argv[i][1] != '-')
		{
			int j;

			if (len >= 3 && argv[i][2] == '=')
			{
				/* Parse the things between the '=' sign. */
				int rightLen = strlen(argv[i] + 3);

				found = __checkArgTypes(&argv[i][1], argTypeCount, argTypes,
										0, &arg);

				if (found)
				{
					if (arg->type == ARG_SETTING)
					{
						if (rightLen > 0)
							*((char **)arg->var) = strdup(argv[i] + 3);
						else
							*((char **)arg->var) = NULL;
					}
					else
					{
						fprintf(stderr,
							_("Warning: Argument `%c' is not a setting.\n"),
							argv[i][1]);
					}
				}
			}
			else
			{
				char doContinue = 0;

				for (j = 1; j < len; j++)
				{
					found = __checkArgTypes(&argv[i][j], argTypeCount,
											argTypes, 0, &arg);

					if (found && arg->type == ARG_SETTING)
					{
						*((char **)arg->var) = argv[i + 1];

						i++;

						doContinue = 1;
						break;
					}
				}

				if (doContinue)
					continue;
			}
		}
		/* Check for '--xyz' opts */
		else if (len > 2 && argv[i][0] == '-')
		{
			char *eq;
			
			if (argv[i][1] != '-')
			{
				INVALID_ARG();
			}

			if ((eq = strchr(argv[i], '=')) != NULL)
			{
				/* Parse the things between the '=' sign. */
				int leftLen  = eq - argv[i] - 2;
				int rightLen = strlen(argv[i]) + leftLen + 2;
				char *left, *right;

				left  = (char *)malloc(leftLen + 1);
				right = (char *)malloc(rightLen +1);

				strncpy(left, argv[i] + 2, leftLen);
				left[leftLen] = '\0';
				
				strncpy(right, argv[i] + 2 + leftLen + 1, rightLen);
				right[rightLen] = '\0';

				found = __checkArgTypes(left, argTypeCount, argTypes,
										1, &arg);

				if (found)
				{
					if (arg->type == ARG_SETTING)
					{
						*((char **)arg->var) = strdup(right);
					}
					else
					{
						fprintf(stderr,
							_("Warning: Argument `%s' is not a setting.\n"),
							left);
					}
				}

				free(left);
				free(right);
			}
			else
			{
				found = __checkArgTypes(argv[i] + 2, argTypeCount, argTypes,
										1, &arg);
			
				if (found && arg->type == ARG_SETTING)
				{
					*((char **)arg->var) = argv[i + 1];
					
					i++;

					continue;
				}
			}
		}
		else
		{
			/* Check for commands */
			if (getCommand() != CMD_UNSPECIFIED)
			{
				/*
				 * A command was specified.
				 * Assume this is a file or something.
				 */
				(*newArgv)[(*newArgc)++] = argv[i];

				continue; /* Skip the 'found' check */
			}
			else
			{
				found = __checkArgTypes(argv[i], argTypeCount, argTypes,
										2, &arg);
			}
		}

		if (found == 0 && newArgc != NULL && newArgv != NULL)
		{
			if (keepUnknownFlags == 1)
			{
				(*newArgv)[(*newArgc)++] = argv[i];
			}
			else
			{
				INVALID_ARG();
			}
		}
	}
}

